# Copyright (C) 2024 by Siisurit e.U., Austria.
# All rights reserved. Distributed under the MIT License.
import argparse
import logging
import sys
from pathlib import Path

import check_done
from check_done.config import (
    CONFIG_BASE_NAME,
    default_config_path,
    map_from_yaml_file_path,
    validate_configuration_info_from_yaml_map,
)
from check_done.done_project_items_info import done_project_items_info
from check_done.warning_checks import warnings_for_done_project_items

logger = logging.getLogger(__name__)

_HELP_DESCRIPTION = (
    'Check that GitHub issues and pull requests in a project board with a status of "Done" are really done.'
)


def check_done_command(arguments=None) -> int:
    result = 1
    try:
        execute(arguments)
        result = 0
    except KeyboardInterrupt:
        logger.error("Interrupted as requested by user.")  # noqa: TRY400
    except Exception:
        logger.exception("Cannot check done project items.")
    return result


def execute(arguments=None):
    parser = _argument_parser()
    args = parser.parse_args(arguments)
    configuration_yaml_path = args.config or default_config_path()
    yaml_map = map_from_yaml_file_path(configuration_yaml_path)
    configuration_info = validate_configuration_info_from_yaml_map(yaml_map)
    done_project_items = done_project_items_info(configuration_info)
    done_project_items_count = len(done_project_items)
    if done_project_items_count == 0:
        logger.info("Nothing to check. Project has no items in the selected project status.")
    else:
        warnings = warnings_for_done_project_items(done_project_items)
        if len(warnings) == 0:
            logger.info(
                f"All project items are correct, "
                f"{done_project_items_count!s} checked in the selected project status. "
            )
        else:
            for warning in warnings:
                logger.warning(warning)


def _argument_parser():
    parser = argparse.ArgumentParser(prog="check_done", description=_HELP_DESCRIPTION)
    parser.add_argument(
        "--config",
        "-c",
        type=Path,
        help=(
            f"Path to configuration file with project URL, authentication details, and possibly other options; "
            f"default: {CONFIG_BASE_NAME}.yaml in the current working directory or any of the above."
        ),
    )
    parser.add_argument("--version", "-v", action="version", version="%(prog)s " + check_done.__version__)
    return parser


def main():  # pragma: no cover
    logging.basicConfig(level=logging.INFO)
    sys.exit(check_done_command())


if __name__ == "__main__":  # pragma: no cover
    main()



# Copyright (C) 2024 by Siisurit e.U., Austria.
# All rights reserved. Distributed under the MIT License.
from html.parser import HTMLParser

from check_done.info import GithubProjectItemType, ProjectItemInfo


class _StopParsingHtml(Exception):
    """Custom exception to stop HTML parsing."""


class _AllTasksCheckedHtmlParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.all_tasks_are_checked = True

    def handle_starttag(self, tag, attrs):
        if tag == "input":
            attr_dict = dict(attrs)
            is_checkbox = attr_dict.get("type") == "checkbox"
            if is_checkbox:
                is_checked = "checked" in attr_dict
                if not is_checked:
                    self.all_tasks_are_checked = False
                    raise _StopParsingHtml


def warnings_for_done_project_items(done_project_items: list[ProjectItemInfo]) -> list[str | None]:
    result = []
    for project_item in done_project_items:
        warning_reasons = [
            warning_reason(project_item)
            for warning_reason in POSSIBLE_WARNINGS
            if warning_reason(project_item) is not None
        ]
        if len(warning_reasons) >= 1:
            warning = sentence_from_project_item_warning_reasons(project_item, warning_reasons)
            result.append(warning)
    return result


def sentence_from_project_item_warning_reasons(project_item: ProjectItemInfo, warning_reasons: list[str]) -> str:
    if len(warning_reasons) >= 3:
        warning_reasons = f"{', '.join(warning_reasons[:-1])}, and {warning_reasons[-1]}"
    elif len(warning_reasons) == 2:
        warning_reasons = f"{', '.join(warning_reasons[:1])} and {warning_reasons[1]}"
    else:
        warning_reasons = warning_reasons[0]
    return (
        f" Project item should {warning_reasons}."
        f" - repository: {project_item.repository.name!r} - item name: '#{project_item.number} {project_item.title}'."
    )


def warning_reason_if_open(project_item: ProjectItemInfo) -> str | None:
    return "be closed" if not project_item.closed else None


def warning_reason_if_unassigned(project_item: ProjectItemInfo) -> str | None:
    return "be assigned" if project_item.assignees.total_count == 0 else None


def warning_reason_if_missing_milestone(project_item: ProjectItemInfo) -> str | None:
    return "have a milestone" if project_item.milestone is None else None


def warning_reason_if_tasks_are_uncompleted(project_item: ProjectItemInfo) -> str | None:
    # TODO#29 Change parsing of tasks to markdown description.
    #  Background: This is less fragile than HTML because even if GitHub changes the names of HTML
    #  types it will still work.
    parser = _AllTasksCheckedHtmlParser()
    try:
        parser.feed(project_item.body_html)
        parser.close()
    except _StopParsingHtml:
        pass  # Stop parsing HTML after first unchecked task.
    return "have all tasks completed" if not parser.all_tasks_are_checked else None


def warning_reason_if_missing_closing_issue_reference_in_pull_request(project_item: ProjectItemInfo) -> str | None:
    is_missing_closing_issue_reference_in_pull_request = (
        len(project_item.closing_issues_references.nodes) == 0
        if project_item.typename is GithubProjectItemType.pull_request
        else False
    )
    return "have a closing issue reference" if is_missing_closing_issue_reference_in_pull_request else None


POSSIBLE_WARNINGS = [
    warning_reason_if_open,
    warning_reason_if_unassigned,
    warning_reason_if_missing_milestone,
    warning_reason_if_tasks_are_uncompleted,
    warning_reason_if_missing_closing_issue_reference_in_pull_request,
]



# Copyright (C) 2024 by Siisurit e.U., Austria.
# All rights reserved. Distributed under the MIT License.
import time

import jwt
import requests
from requests import Session

from check_done.graphql import HttpBearerAuth

_SECONDS_PER_MINUTE = 60
_ISSUED_AT = int(time.time())
_EXPIRES_AT = int(time.time()) + (10 * _SECONDS_PER_MINUTE)


class AuthenticationError(Exception):
    """Error raised due to failed JWT authentication process."""


def resolve_organization_access_token(organization_name: str, github_app_id: str, github_app_private_key: str) -> str:
    """
    Generates the necessary access token for an organization from the installed GitHub app instance in said organization
    """
    jwt_token = generate_jwt_token(github_app_id, github_app_private_key)
    session = requests.Session()
    session.headers = {"Accept": "application/vnd.github+json"}
    session.auth = HttpBearerAuth(jwt_token)
    try:
        github_app_installation_id = resolve_github_app_installation_id(session, organization_name)
        result = resolve_access_token_from_github_app_installation_id(session, github_app_installation_id)
    except Exception as error:
        raise AuthenticationError(
            f"Cannot resolve organization access token from JWT authentication process: {error}"
        ) from error
    return result


def generate_jwt_token(github_app_id: str, github_app_private_key: str) -> str:
    """Generates a JWT token for authentication with GitHub."""
    try:
        payload = {
            "exp": _EXPIRES_AT,
            "iat": _ISSUED_AT,
            "iss": github_app_id,
        }
        return jwt.encode(payload, github_app_private_key, algorithm="RS256")
    except Exception as error:
        raise AuthenticationError(f"Cannot generate JWT token: {error}") from error


def resolve_github_app_installation_id(session: Session, organization_name: str) -> str:
    """Fetches the installation ID for the organization."""
    response = session.get(f"https://api.github.com/orgs/{organization_name}/installation")

    if response.status_code == 200 and response.json().get("id") is not None:
        return response.json().get("id")
    raise AuthenticationError(
        f"Could not retrieve installation ID: status={response.status_code} - response_text={response.text}"
    )


def resolve_access_token_from_github_app_installation_id(session: Session, installation_id: str) -> str:
    """Retrieves the access token using the installation ID."""
    response = session.post(f"https://api.github.com/app/installations/{installation_id}/access_tokens")
    if response.status_code == 201 and response.json().get("token") is not None:
        return response.json().get("token")
    raise AuthenticationError(
        f"Could not retrieve access token: status={response.status_code} - response_text={response.text}"
    )



# Copyright (C) 2024 by Siisurit e.U., Austria.
# All rights reserved. Distributed under the MIT License.
import re
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import Any

from pydantic import BaseModel
from requests import HTTPError, Response, Session
from requests.auth import AuthBase

from check_done.info import (
    QueryInfo,
)

GRAPHQL_ENDPOINT = "https://api.github.com/graphql"
_MAX_ENTRIES_PER_PAGE = 100
_PATH_TO_QUERIES = Path(__file__).parent / "queries"


class GraphQlError(Exception):
    pass


class HttpBearerAuth(AuthBase):
    # Source:
    # <https://stackoverflow.com/questions/29931671/making-an-api-call-in-python-with-an-api-that-requires-a-bearer-token>
    def __init__(self, token):
        self.token = token

    def __call__(self, request):
        request.headers["Authorization"] = "Bearer " + self.token
        return request


@lru_cache
def minimized_graphql(graphql_query: str) -> str:
    single_spaced_query = re.sub(r"\s+", " ", graphql_query)
    formatted_query = re.sub(r"\s*([{}()\[\]:,])\s*", r"\1", single_spaced_query)
    result = formatted_query.strip()
    if result == "":
        raise GraphQlError("GraphQL query is empty.")
    return result


def _graphql_query(item_name: str) -> str:
    query_path = _PATH_TO_QUERIES / f"{item_name}.graphql"
    with query_path.open(encoding="utf-8") as query_file:
        query_text = query_file.read()
    return minimized_graphql(query_text)


class GraphQlQuery(Enum):
    ORGANIZATION_PROJECTS = _graphql_query("organization_projects")
    USER_PROJECTS = _graphql_query("user_projects")
    PROJECT_SINGLE_SELECT_FIELDS = _graphql_query("project_single_select_fields")
    PROJECT_V2_ITEMS = _graphql_query("project_v2_items")

    @staticmethod
    def query_for(name: str):
        assert name is not None
        assert name.upper() == name
        return GraphQlQuery[name].value


def query_infos(
    base_model: type[BaseModel],
    query_name: str,
    session: Session,
    project_owner_name: str,
    project_id: str | None = None,
) -> list:
    result = []
    variables = {"login": project_owner_name, "maxEntriesPerPage": _MAX_ENTRIES_PER_PAGE}
    if project_id is not None:
        variables["projectId"] = project_id
    after = None
    has_more_pages = True
    while has_more_pages:
        query = GraphQlQuery.query_for(query_name)
        if after is not None:
            variables["after"] = after
        json_payload_map = {
            "variables": variables,
            "query": query,
        }
        response = session.post(GRAPHQL_ENDPOINT, json=json_payload_map)
        response_map = checked_graphql_data_map(response)
        response_info = base_model(**response_map)
        query_info = query_info_from_response_info(response_info)
        nodes_info = query_info.nodes
        page_info = query_info.page_info
        after = page_info.endCursor
        has_more_pages = page_info.hasNextPage
        result.extend(nodes_info)
    return result


def query_info_from_response_info(base_model: BaseModel) -> QueryInfo:
    if isinstance(base_model, QueryInfo):
        return base_model
    for model_field_name in base_model.model_fields_set:
        field_model = getattr(base_model, model_field_name)
        if isinstance(field_model, BaseModel):
            result = query_info_from_response_info(field_model)
            if result is not None:
                return result
    raise ValueError(
        f"Could not find fields matching the {QueryInfo.__name__} model with fields "
        f"`{QueryInfo.model_fields.keys()}` in: {base_model}"
    )


def checked_graphql_data_map(response: Response) -> dict[str, Any | None]:
    try:
        response.raise_for_status()
    except HTTPError as error:
        raise GraphQlError(error) from error
    response_map = response.json()
    if not isinstance(response_map, dict):
        raise GraphQlError(f"GraphQL response must be a map but is: {response_map}.")
    errors = response_map.get("errors")
    if errors is not None:
        raise GraphQlError(f'{errors[0]["message"]}; details: {errors}.')
    result = response_map.get("data")
    if result is None:
        raise GraphQlError(f"GraphQL result must include data but only has: {sorted(response_map.keys())}.")
    if not isinstance(result, dict):
        raise GraphQlError(f"GraphQL data must be a map but is: {result}.")
    return result



# Copyright (C) 2024 by Siisurit e.U., Austria.
# All rights reserved. Distributed under the MIT License.
import os
import re
import string
from enum import StrEnum
from pathlib import Path
from typing import Any

import yaml
from dotenv import load_dotenv
from pydantic import Field, field_validator, model_validator
from pydantic.dataclasses import dataclass

load_dotenv()

CONFIG_BASE_NAME = ".check_done"
_GITHUB_ORGANIZATION_NAME_AND_PROJECT_NUMBER_URL_REGEX = re.compile(
    r"https://github\.com/orgs/(?P<organization_name>[a-zA-Z0-9\-]+)/projects/(?P<project_number>[0-9]+).*"
)
_GITHUB_USER_NAME_AND_PROJECT_NUMBER_URL_REGEX = re.compile(
    r"https://github\.com/users/(?P<user_name>[a-zA-Z0-9\-]+)/projects/(?P<project_number>[0-9]+).*"
)


class ProjectOwnerType(StrEnum):
    User = "users"
    Organization = "orgs"


@dataclass
class ConfigurationInfo:
    project_url: str

    # Required for user owned project
    personal_access_token: str | None = None

    # Required for organization owned project
    github_app_id: str | None = None
    github_app_private_key: str | None = None

    # Optional
    project_status_name_to_check: str | None = None

    # Fields computed during initialization
    is_project_owner_of_type_organization: bool = Field(init=False)
    project_number: int = Field(init=False)
    project_owner_name: str = Field(init=False)

    @field_validator(
        "project_url",
        "project_status_name_to_check",
        "personal_access_token",
        "github_app_id",
        "github_app_private_key",
        mode="before",
    )
    def value_from_env(cls, value: Any | None):
        stripped_value = value.strip()
        result = (
            resolved_environment_variables(value, fail_on_missing_envvar=False)
            if stripped_value.startswith("${") and stripped_value.endswith("}")
            else stripped_value
        )
        return result

    @model_validator(mode="after")
    def validate_authentication_and_set_project_details(self):
        self.project_owner_name, self.project_number, self.is_project_owner_of_type_organization = (
            github_project_owner_name_and_project_number_and_is_project_owner_of_type_organization_from_url_if_matches(
                self.project_url
            )
        )

        has_user_authentication = (
            self.personal_access_token is not None and not self.is_project_owner_of_type_organization
        )
        has_organizational_authentication = (
            self.github_app_id is not None
            and self.github_app_private_key is not None
            and self.is_project_owner_of_type_organization
        )
        if not has_user_authentication ^ has_organizational_authentication:
            raise ValueError("A user or an organization authentication method must be configured.")
        return self


def validate_configuration_info_from_yaml_map(yaml_map: dict) -> ConfigurationInfo:
    return ConfigurationInfo(**yaml_map)


def github_project_owner_name_and_project_number_and_is_project_owner_of_type_organization_from_url_if_matches(
    url: str,
) -> tuple[str, int, bool]:
    organization_name_and_project_number_match = _GITHUB_ORGANIZATION_NAME_AND_PROJECT_NUMBER_URL_REGEX.match(url)
    if organization_name_and_project_number_match is None:
        user_name_and_project_number_match = _GITHUB_USER_NAME_AND_PROJECT_NUMBER_URL_REGEX.match(url)
        if organization_name_and_project_number_match is None and user_name_and_project_number_match is None:
            raise ValueError(f"Cannot parse GitHub organization or user name, and project number from URL: {url}.")
        project_owner_name = user_name_and_project_number_match.group("user_name")
        project_number = int(user_name_and_project_number_match.group("project_number"))
        is_project_owner_of_type_organization = False
    else:
        project_owner_name = organization_name_and_project_number_match.group("organization_name")
        project_number = int(organization_name_and_project_number_match.group("project_number"))
        is_project_owner_of_type_organization = True
    return project_owner_name, project_number, is_project_owner_of_type_organization


def resolved_environment_variables(value: str, fail_on_missing_envvar=True) -> str:
    try:
        result = string.Template(value).substitute(os.environ)
    except KeyError as error:
        if fail_on_missing_envvar:
            raise ValueError(f"Cannot resolve {value}: environment variable must be set: {error!s}") from error
        result = None
    except ValueError as error:
        raise ValueError(f"Cannot resolve {value!r}: {error}.") from error
    return result


def map_from_yaml_file_path(configuration_path: Path) -> dict:
    try:
        with open(configuration_path) as configuration_file:
            result = yaml.safe_load(configuration_file)
            if result is None:
                raise ValueError(
                    f"The check_done configuration yaml file is empty. Path: {configuration_path}"
                ) from None
    except FileNotFoundError:
        raise FileNotFoundError(f"Cannot find check_done configuration yaml file: {configuration_path}") from None
    return result


def default_config_path() -> Path:
    initial_config_folder = Path().absolute()
    config_folder = initial_config_folder
    previous_config_folder = None
    result = None
    while result is None and config_folder != previous_config_folder:
        config_paths_to_check = [(config_folder / CONFIG_BASE_NAME).with_suffix(suffix) for suffix in [".yaml", ".yml"]]
        result = next((path for path in config_paths_to_check if path.is_file()), None)
        if result is None:
            previous_config_folder = config_folder
            config_folder = config_folder.parent
    if result is None:
        raise FileNotFoundError(
            f"Cannot find configuration file by looking for {CONFIG_BASE_NAME}.yaml "
            f"and traversing upwards starting in {initial_config_folder}"
        )
    return result



# Copyright (C) 2024 by Siisurit e.U., Austria.
# All rights reserved. Distributed under the MIT License.
import logging

import requests

from check_done.config import ConfigurationInfo
from check_done.graphql import GraphQlQuery, HttpBearerAuth, query_infos
from check_done.info import (
    NodeByIdInfo,
    ProjectItemInfo,
    ProjectOwnerInfo,
    ProjectV2ItemNode,
    ProjectV2Node,
    ProjectV2SingleSelectFieldNode,
)
from check_done.organization_authentication import resolve_organization_access_token

_GITHUB_PROJECT_STATUS_FIELD_NAME = "Status"
logger = logging.getLogger(__name__)


def done_project_items_info(configuration_info: ConfigurationInfo) -> list[ProjectItemInfo]:
    project_owner_name = configuration_info.project_owner_name

    is_project_owner_of_type_organization = configuration_info.is_project_owner_of_type_organization
    if is_project_owner_of_type_organization:
        github_app_id = configuration_info.github_app_id
        github_app_private_key = configuration_info.github_app_private_key
        access_token = resolve_organization_access_token(project_owner_name, github_app_id, github_app_private_key)
    else:
        access_token = configuration_info.personal_access_token

    with requests.Session() as session:
        session.headers = {"Accept": "application/vnd.github+json"}
        session.auth = HttpBearerAuth(access_token)

        project_query_name = (
            GraphQlQuery.ORGANIZATION_PROJECTS.name
            if is_project_owner_of_type_organization
            else GraphQlQuery.USER_PROJECTS.name
        )
        project_infos = query_infos(ProjectOwnerInfo, project_query_name, session, project_owner_name)

        project_number = configuration_info.project_number
        project_id = matching_project_id(project_infos, project_number, project_owner_name)
        project_single_select_field_infos = query_infos(
            NodeByIdInfo, GraphQlQuery.PROJECT_SINGLE_SELECT_FIELDS.name, session, project_owner_name, project_id
        )

        project_item_infos = query_infos(
            NodeByIdInfo, GraphQlQuery.PROJECT_V2_ITEMS.name, session, project_owner_name, project_id
        )

    project_status_name_to_check = configuration_info.project_status_name_to_check
    project_status_option_id = matching_project_status_option_id(
        project_single_select_field_infos,
        project_status_name_to_check,
        project_number,
        project_owner_name,
    )
    result = filtered_project_item_infos_by_done_status(project_item_infos, project_status_option_id)
    return result


def matching_project_id(project_infos: list[ProjectV2Node], project_number: int, project_owner_name: str) -> str:
    try:
        return next(str(project_info.id) for project_info in project_infos if project_info.number == project_number)
    except StopIteration:
        raise ValueError(
            f"Cannot find a project with number '{project_number}', owned by '{project_owner_name}'."
        ) from None


def matching_project_status_option_id(
    project_single_select_field_infos: list[ProjectV2SingleSelectFieldNode],
    project_status_name_to_check: str | None,
    project_number: int,
    project_owner_name: str,
) -> str:
    try:
        status_options = next(
            field_info.options
            for field_info in project_single_select_field_infos
            if field_info.name == _GITHUB_PROJECT_STATUS_FIELD_NAME
        )
    except StopIteration:
        raise ValueError(
            f"Cannot find a project status selection field in the GitHub project with number `{project_number}` "
            f"owned by `{project_owner_name}`."
        ) from None
    if project_status_name_to_check:
        logger.info(f"Checking project items with status matching: '{project_status_name_to_check}'")
        try:
            result = next(
                status_option.id
                for status_option in status_options
                if project_status_name_to_check in status_option.name
            )
        except StopIteration:
            project_status_options_names = [
                option.name for field_info in project_single_select_field_infos for option in field_info.options
            ]
            raise ValueError(
                f"Cannot find the project status matching name {project_status_name_to_check!r} "
                f"in the GitHub project with number '{project_number}' owned by {project_owner_name!r}. "
                f"Available options are: {project_status_options_names!r}"
            ) from None
    else:
        last_status_option = status_options[-1]
        logger.info(f"Checking project items with the last project status selected: {last_status_option.name!r}.")
        result = last_status_option.id
    return result


def filtered_project_item_infos_by_done_status(
    project_item_infos: list[ProjectV2ItemNode],
    project_status_option_id: str,
) -> list[ProjectItemInfo]:
    result = []
    for project_item_info in project_item_infos:
        has_project_status_option = (
            project_item_info.field_value_by_name is not None
            and project_item_info.field_value_by_name.option_id == project_status_option_id
        )
        if has_project_status_option:
            result.append(project_item_info.content)
    return result



# Copyright (C) 2024 by Siisurit e.U., Austria.
# All rights reserved. Distributed under the MIT License.
from enum import StrEnum
from typing import Any

from pydantic import AliasChoices, BaseModel, ConfigDict, Field, NonNegativeInt, field_validator


class NodesTypeName(StrEnum):
    CustomField = "CustomField"
    MultiUserIssueCustomField = "MultiUserIssueCustomField"
    PeriodProjectCustomField = "PeriodProjectCustomField"
    PeriodIssueCustomField = "PeriodIssueCustomField"
    ProjectTimeTrackingSettings = "ProjectTimeTrackingSettings"
    SingleUserIssueCustomField = "SingleUserIssueCustomField"


class ProjectItemState(StrEnum):
    CLOSED = "CLOSED"
    OPEN = "OPEN"


class GithubProjectItemType(StrEnum):
    issue = "Issue"
    pull_request = "PullRequest"


class _EmptyDict(BaseModel):
    model_config = ConfigDict(extra="forbid")


class PageInfo(BaseModel):
    endCursor: str
    hasNextPage: bool


class QueryInfo(BaseModel):
    """The nested content of the query and its pagination info in a paginated GraphQL query with nodes"""

    nodes: list[Any]
    page_info: PageInfo = Field(alias="pageInfo")

    @field_validator("nodes", mode="after", check_fields=True)
    def resolve_nodes(cls, nodes: list[Any]):
        validated_nodes = []
        for node in nodes:
            node_type = node.get("__typename")
            if node.get("__typename") in _NODE_TYPE_NAME_TO_INFO_CLASS_MAP:
                node_model = _NODE_TYPE_NAME_TO_INFO_CLASS_MAP.get(node_type)
                validated_nodes.append(node_model(**node))
        return validated_nodes


class ProjectV2Node(BaseModel):
    id: str
    number: NonNegativeInt
    typename: str = Field(alias="__typename")
    fields: QueryInfo | None = None
    items: QueryInfo | None = None


class ProjectV2Options(BaseModel):
    id: str
    name: str


class ProjectV2SingleSelectFieldNode(BaseModel):
    id: str
    name: str
    typename: str = Field(alias="__typename")
    options: list[ProjectV2Options]


class ProjectV2ItemProjectStatusInfo(BaseModel):
    status: str
    option_id: str = Field(alias="optionId")


class RepositoryInfo(BaseModel):
    name: str


class AssigneesInfo(BaseModel):
    total_count: NonNegativeInt = Field(alias="totalCount")


class MilestoneInfo(BaseModel):
    id: str


class LinkedProjectItemNode(BaseModel):
    number: NonNegativeInt
    title: str


class LinkedProjectItemInfo(BaseModel):
    nodes: list[LinkedProjectItemNode]


# NOTE: For simplicity, both issues and pull requests are treated the same under a generic "project item" type.
#  Since all their underlying properties needed for checking, are mostly identical.
class ProjectItemInfo(BaseModel):
    """A generic type representing both issues and pull requests in a project board."""

    # Shared between Issues and Pull Requests.
    typename: GithubProjectItemType = Field(alias="__typename")
    assignees: AssigneesInfo
    body_html: str = Field(alias="bodyHTML", default=None)
    closed: bool
    number: NonNegativeInt
    repository: RepositoryInfo
    milestone: MilestoneInfo | None
    title: str

    # Only set for pull requests
    closing_issues_references: LinkedProjectItemInfo = Field(alias="closingIssuesReferences", default=None)


class ProjectV2ItemNode(BaseModel):
    content: ProjectItemInfo | _EmptyDict = None
    field_value_by_name: ProjectV2ItemProjectStatusInfo | None = Field(alias="fieldValueByName", default=None)
    typename: str = Field(alias="__typename")


class NodeByIdInfo(BaseModel):
    node: ProjectV2Node


class _ProjectsV2Info(BaseModel):
    projects_v2: QueryInfo = Field(alias="projectsV2")


class ProjectOwnerInfo(BaseModel):
    """
    A generic type representing the project owner whether they are an organization or user.
    The difference between organization and user is irrelevant in check_done's case,
    as we are only using the project owner to retrieve the projects ID's.
    """

    project_owner: _ProjectsV2Info = Field(validation_alias=AliasChoices("organization", "user"))


class _NodeTypeName(StrEnum):
    ProjectV2 = "ProjectV2"
    ProjectV2SingleSelectField = "ProjectV2SingleSelectField"
    ProjectV2Item = "ProjectV2Item"


_NODE_TYPE_NAME_TO_INFO_CLASS_MAP = {
    _NodeTypeName.ProjectV2.value: ProjectV2Node,
    _NodeTypeName.ProjectV2SingleSelectField.value: ProjectV2SingleSelectFieldNode,
    _NodeTypeName.ProjectV2SingleSelectField.ProjectV2Item: ProjectV2ItemNode,
}
